#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Tech - Sérialisation
# 
# La sérialisation répond à un problème simple : comment échanger des données complexes autres que des tableaux ?
# 
# Si l'énoncé est simple, la solution ne l'est pas toujours. Il est assez facile d'échanger des données qui se présentent sous la forme d'un tableau, d'un texte, d'un nombre mais comment échanger un assemblage de données hétérogènes ? La **sérialisation** désigne un méanisme qui permet de permet de représenter un assemblage de données en un seul tableau de caractères. La **désérialisation** désigne le mécanisme inverse qui consiste à reconstruire les données initiales à partir de ce tableau de caractères.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Notion de stream ou flux
# 
# Un stream en informatique définit une façon de parcourir une séquence d'octets. Un fichier est un stream : on écrit les octets ou caractères les uns après les autres, chaque nouveau caractère est ajouté à la fin. Lors de la lecture, on procède de même en lisant les caractères du début à la fin. Dans un stream, on ne revient jamais en arrière, on lit toujours le caractère suivant.
# 
# Les streams sont optimisés pour ce type de lecture et d'écriture, ils sont très lents lorsqu'il s'agit d'aller lire ou écrire des caractères de façon non séquentielle.
# 
# Pour faire des calculs mathématiques, il faut pouvoir accéder à tout moment à n'importe quel élément de la matrice. L'utilisation d'un *stream* est contre-indiquée. En revanche, ils sont très adaptés à la lecture et l'écriture de fichiers. Ils sont également utilisés pour communiquer des données, lorsqu'un ordinateur envoie des données à un autre ordinateur.

# In[2]:


import math
from io import StringIO

st = StringIO()
st.write("pi=")
st.write(str(math.pi))
st.write(";")
value = st.getvalue()
print(value)


# In[3]:


st = StringIO("pi=3.141592653589793;")
while text := st.read(1):
    print(text)


# In[4]:


def f1(text):
    st = StringIO()
    for t in text:
        st.write(t)
        st.write(";")
    value = st.getvalue()
    return value

def f2(text):
    s = ""
    for t in text:
        s += t + ";"
    return s

data = ["petit", "essai", "de", "comparaison"] * 300
f1(data)[:100], f2(data)[:100]


# In[5]:


get_ipython().run_line_magic('timeit', 'f1(data)')


# In[6]:


get_ipython().run_line_magic('timeit', 'f2(data)')


# ### Q1: quelle est la fonction la plus rapide ?
# 
# Il vaut mieux faire varier la longueur de la liste `data` avant de répondre.

# ## Format JSON
# 
# Le format [JSON](https://fr.wikipedia.org/wiki/JavaScript_Object_Notation) est le format le plus répandu sur Internet. C'est un assemblage récursif de listes et de dictionnaires. Chaque conteneur peut contenir des listes, des dictionnaires, des nombres, des chaînes de caractères.
# 
# Il est possible de télécharger tout Wikipédia au format JSON : [Wikidata:Téléchargement de la base de données](https://www.wikidata.org/wiki/Wikidata:Database_download/fr).

# In[7]:


data = {
    "nom": "magoo", 
    "naissance": 1949, 
    "creator": ["Millard Kaufman", "John Hubley"],
    "cartoons": [
        {"title": "Les Aventures célèbres de Monsieur Magoo", "durée": 5},
        {"title": "Quoi de neuf Mr. Magoo ?", "durée": 10}
    ]
}
data


# Le langage python propose une librairie standard [json](https://docs.python.org/3/library/json.html) pour manipuler les informations. Et comme c'est d'un usage fréquent, il existe d'autres options plus rapides [ujson](https://github.com/ultrajson/ultrajson), [simplejson](https://github.com/simplejson/simplejson), [ijson](https://github.com/ICRAR/ijson), ...
# 
# La page [Index of /wikidatawiki/entities/](https://dumps.wikimedia.org/wikidatawiki/entities/) contient des fichiers json issues de wikipedia. Le fichier [latest-lexemes-sample.json](https://raw.githubusercontent.com/sdpython/ensae_teaching_cs/master/_doc/notebooks/td1a_home/latest-lexemes-sample.json) contient les premières lignes de ``latest-lexemes.json.bz2``.

# ### Q1. lire du json
# 
# Télécharger et lire ce [fichier](https://raw.githubusercontent.com/sdpython/ensae_teaching_cs/master/_doc/notebooks/td1a_home/latest-lexemes-sample.json) avec la libraire [json](https://docs.python.org/3/library/json.html).

# In[8]:


from urllib.request import urlopen
url = "https://raw.githubusercontent.com/sdpython/ensae_teaching_cs/master/_doc/notebooks/td1a_home/latest-lexemes-sample.json"
with urlopen(url) as f:
    text = f.read().decode("utf-8")
print(text[:150] + "...")


# In[9]:





# ### Q2: écrire du json
# 
# Modifier les données et les écrire de nouveau sur disque.

# In[10]:





# ### Q3: gros json
# 
# Le dump de la version anglaise de wikipedia fait plus de 100 Go (en version compressée). Il tient sur disque mais pas en mémoire. Comment faire pour le lire malgré tout ? Quelques lignes pour vous données des idées... Les plus courageux utiliseront la librairie [ijson](https://github.com/ICRAR/ijson) ou [orjson](https://github.com/ijl/orjson).

# In[11]:


with open("dummy.json", "w", encoding="utf-8") as f:
    f.write(text)
    
with open("dummy.json", "r", encoding="utf-8") as f:
    for i, line in enumerate(f):
        if i > 2:
            break
        print(line)


# ## XML
# 
# Le [XML](https://fr.wikipedia.org/wiki/Extensible_Markup_Language) était utilisé avant le format json. Il permet de faire la même chose, sérialiser, mais est plus verbeux. Il a été abandonné car le résultat est plus long qu'avec le format json.

# In[12]:


from dict2xml import dict2xml
print(dict2xml(data))


# ## pickle
# 
# Le format [JSON](https://fr.wikipedia.org/wiki/JavaScript_Object_Notation) a un inconvénient majeur : il impose la conversion des données au format texte, en particulier les nombres. Chaque nombre doit être converti en chaînes de caractères et réciproquement. Pourquoi ne pas garder la représentation binaire des nombres tels qu'ils sont utilisés en mémoire ?
# 
# C'est l'objectif du module [pickle](https://docs.python.org/3/library/pickle.html). Comme il n'y pas de conversion au format texte et qu'il s'agit de recopier la mémoire sur disque en un seule, cette sérialisation s'applique à tout objet python. Elle n'est pas restreinte aux dictionnaires et aux listes.

# ### Q1: comparer le temps de sérialisation entre pickle et json
# 
# On pourra utiliser les données json récupérées ci-dessus.

# In[13]:





# ### Q2: comparer le temps de désérialisation entre pickle et json
# 
# Même exercice en sens inverse.

# In[14]:





# ### Peut-on tout sérialiser ?
# 
# La plupart des objets contenant des données peuvent être sérialisées, les listes, ldes dictionnaires, les matrices (numpy), les dataframes)... Il n'est pas possible de sérialiser les fonctions à moins d'utiliser des librairies comme [cloudpickle](https://github.com/cloudpipe/cloudpickle) ou [dill](https://github.com/uqfoundation/dill).
# 
# La sérialisation fonctionne de façon implicite avec toutes les classes python à l'exception de celles définies en C++. Pour celles-ci, il faudra coder explicitement la sérialisation et la désérialisation. Pour cela il faut redéfinir les méthodes [getstate et_setstate](http://www.xavierdupre.fr/app/teachpyx/helpsphinx/notebooks/serialisation_examples.html?highlight=__getstate__#reduire-la-taille).
# 
# Il reste une contrainte majeure à cette sérialisation, elle dépend de la version du langage et de chaque extension. Sérialisation avec python 3.7 et désérialisation avec python 3.10 a peu de chance de fonctionner.

# In[15]: